"use strict";

var _interopRequireDefault = require("@babel/runtime-corejs3/helpers/interopRequireDefault");

var _Object$defineProperty = require("@babel/runtime-corejs3/core-js-stable/object/define-property");

_Object$defineProperty(exports, "__esModule", {
  value: true
});

exports.getAddressToContract = getAddressToContract;
exports.getTransactionParams = getTransactionParams;

var _indexOf = _interopRequireDefault(require("@babel/runtime-corejs3/core-js-stable/instance/index-of"));

var _filter = _interopRequireDefault(require("@babel/runtime-corejs3/core-js-stable/instance/filter"));

var _map = _interopRequireDefault(require("@babel/runtime-corejs3/core-js-stable/instance/map"));

var _forEach = _interopRequireDefault(require("@babel/runtime-corejs3/core-js-stable/instance/for-each"));

var _embarkUtils = require("embark-utils");

function getAddressToContract(contractsList, addressToContract) {
  if (!contractsList) {
    return addressToContract;
  }

  (0, _forEach.default)(contractsList).call(contractsList, contract => {
    var _context, _context2;

    if (!contract.deployedAddress) {
      return;
    }

    const address = contract.deployedAddress.toLowerCase();

    if (addressToContract[address]) {
      return;
    }

    const funcSignatures = {};
    (0, _map.default)(_context = (0, _filter.default)(_context2 = contract.abiDefinition).call(_context2, func => func.type === "function")).call(_context, func => {
      var _context3;

      const name = `${func.name}(${func.inputs ? (0, _map.default)(_context3 = func.inputs).call(_context3, input => input.type).join(",") : ""})`;
      funcSignatures[(0, _embarkUtils.sha3)(name).substring(0, 10)] = {
        abi: func,
        functionName: func.name,
        name
      };
    });
    addressToContract[address] = {
      functions: funcSignatures,
      name: contract.className,
      silent: contract.silent
    };
  });
  return addressToContract;
}

function getTransactionParams(contract, transactionInput) {
  const func = contract.functions[transactionInput.substring(0, 10)];
  const functionName = func ? func.functionName : "UNKNOWN";
  let paramString = "";

  if (func && func.abi && func.abi.inputs) {
    var _context4;

    const decodedParameters = (0, _embarkUtils.decodeParams)(func.abi.inputs, transactionInput.substring(10));
    (0, _forEach.default)(_context4 = func.abi.inputs).call(_context4, input => {
      var _context5;

      const quote = (0, _indexOf.default)(_context5 = input.type).call(_context5, "int") === -1 ? '"' : "";
      paramString += quote + decodedParameters[input.name] + quote + ", ";
    });
    paramString = paramString.substring(0, paramString.length - 2);
  }

  return {
    functionName,
    paramString
  };
}
//# sourceMappingURL=transactionUtils.js.map